import shelve
import time
import queue
import random
import string
import math
import logging
import requests
from typing import Optional, Text, Union
from datetime import datetime
from .exceptions import DataTypeError
from collections import namedtuple

ColumnDescription = namedtuple("ColumnDescription", ["col_name", "data_type", "ordinal_position", "is_nullable", "precision"])
logger = logging.getLogger(__name__)

sql_datatypes_map = {
        "postgres": {   
                        "boolean": bool,
                        "character varying": str,
                        "varchar": str,
                        "character": str,
                        "char": str,
                        "text": str,
                        "real": float,
                        "double precision": float,
                        "decimal": float,
                        "numeric": float,
                        "money": float,
                        "integer": int,
                        "smallint": int,
                        "bigint": int,
                        "smallserial": int,
                        "serial": int,
                        "bigserial": int,
                        "timestamp": datetime,
                        "timestamp without time zone": datetime,
                        "interval": str,
                        "cidr": str,
                        "inet": str,
                        "macaddr": str,
                        "macaddr8": str,
                        "USER-DEFINED": str,
                    }
    }

def generate_rand_name() -> str:
    """Generate a random name of the form "export_XXXXXX" where XXXXXX are
    6 random characters.

    Returns:
        str: the generated name
    """
    suf = "".join(random.choices(string.ascii_uppercase + string.digits, k=6))
    return f"exporters_{suf}"

def is_able_to_connect(url: Optional[Text] = None) -> bool:
    """Runs a HTTP GET request to the url.

    Args:
        url (str): an URL (with its schema)

    Returns:
        bool: if a ConnectionError or Timeout are raised, returns `False`.
            `True`, otherwise.
    """
    if url is None:
        url = "https://www.google.com"
    try:
    	_ = requests.get(url, timeout=.5)
    	return True
    except (requests.ConnectionError, requests.Timeout):
    	return False

def _stringify_sql(value):
    if isinstance(value, str):
        return f"'{value}'"
    elif value is None:
        return "NULL"
    else:
        return str(value)

def _transform_data_for_sql_query(data: Union[dict, tuple]):
    generator = enumerate(data) if isinstance(data, tuple) else data.items()
    for k, v in generator:
        if isinstance(v, bool):
            data[k] = "true" if v else "false"
        if isinstance(v, datetime):
            data[k] = str(v)
        if isinstance(v, float):
            if v == math.inf:
                data[k] = 'Infinity'
            elif v == -math.inf:
                data[k] = '-Infinity'
            elif math.isnan(v):
                data[k] = 'NaN'
        if isinstance(v, list):
            data[k] = str(v).replace("[", "{").replace("]", "}")

    columns = None
    if isinstance(data, dict):
        columns = ",".join(data.keys())

    values = ",".join([ f"'{val}'" if isinstance(val, str) else _stringify_sql(val) for val in data.values()])
    values = values.replace("None", "NULL")
    return columns, values

def _describe_postgres_column(col: tuple):
    """Input column is a tuple of the form:
        (table_name, column_name, data_type, ordinal_position, is_nullable, character_maximum_length, numeric_precision, datetime_precision)
    """
    if col[5]:
        precision = col[5]
    elif col[6]:
        precision = col[6]
    else:
        precision = col[7]

    try:
        data_type = sql_datatypes_map["postgres"][col[2]]
    except KeyError:
        logger.error(f"Data type not found in mapping: {col[2]}")
        data_type = None
    
    ordinal_position = col[3]
    is_nullable = True if col[4] == 'YES' else False

    return ColumnDescription(col[1], data_type, ordinal_position, is_nullable, precision)

def _validate_data_for_sql_table(data: dict, table: dict):
    """Validates data based on a table's schema.
    `table` is a dictionary where keys are column names and values are of the type 
    `resilient_exporters.utils.ColumnDescription`.
    """
    for key, val in data.items():
        if val is None:
            if not table[key].is_nullable:
                raise DataTypeError(self, f"Column '{key}' is not nullable, but \
                                            value provided is None.")
        elif not isinstance(val, table[key].data_type):
            raise DataTypeError(self, f"Invalid data type for '{key}'.")
        elif isinstance(val, str):
            if isinstance(table[key].precision, int):
                if len(val) > table[key].precision:
                    raise DataTypeError(self, f"String of chars too long for '{key}'. \
                                                It must be {table[key][3]} chars maximum.")
    return

class _DataStore:
    __instantiated = 0
    __used_filenames = []

    def __new__(cls, *args, **kwargs):
        if "shelf_filename" in kwargs.keys():
            if kwargs["shelf_filename"] in cls.__used_filenames:
                raise ValueError(f"File {kwargs['db_filename']} is already \
                                 being used.")
        return super(_DataStore, cls).__new__(cls)

    def __init__(self,
                 use_memory: bool = True,
                 shelf_filename: Optional[Text] = None,
                 max_size: int = 100 * 100 * 100,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__use_memory = use_memory

        self.__filename = generate_rand_name() if shelf_filename is None \
                                               else shelf_filename
        self.__used_filenames.append(self.__filename)

        self.__size = 0
        self.max_size = max_size

        self.__queue = queue.Queue() if use_memory else None
        self.__shelf = None if use_memory else shelve.open(self.__filename)

    @property
    def size(self):
        return self.__size

    @property
    def use_memory(self) -> bool:
        return self.__use_memory

    @use_memory.setter
    def use_memory(self, new_val: bool) -> bool:
        if self.__use_memory and new_val == False:
            self.export_queue_to_shelf()
            self.__use_memory = new_val
            self.__queue = None
        elif self.__use_memory == False and new_val:
            self.__queue = self.import_queue_from_shelf()
            self.__use_memory = new_val
            self.__shelf.close()
            self.__shelf = None
        else:
            print(f"WARNING - use_memory is already set to {new_val}")
        return new_val

    def put(self, data) -> bool:
        if self.size >= self.max_size:
            # Cannot save more data
            return False
        if self.use_memory:
            self.__put_in_memory(data)
        else:
            self.__put_in_shelf(data)
        self.__size += 1
        return True

    def __put_in_memory(self, data):
        self.__queue.put(data)

    def __put_in_shelf(self, data):
        self.__shelf[str(time.time())] = data

    def get(self):
        if self.size <= 0:
            raise Exception("No saved data left.")
        if self.use_memory:
            data = self.__get_from_memory()
        else:
            data = self.__get_from_shelf()
        self.__size -= 1
        return data

    def __get_from_memory(self):
        return self.__queue.get()

    def __get_from_shelf(self):
        generator = iter(self.__shelf.keys())
        try:
            key = next(generator)
        except StopIteration:
            return None
        else:
            res = self.__shelf[key].copy()
            del self.__shelf[key]
            return res

    def export_queue_to_shelf(self):
        self.__shelf = shelve.open(self.__filename, "n")
        while not self.__queue.empty():
            data = self.__queue.get()
            self.__shelf[str(time.time())] = data

    def import_queue_from_shelf(self):
        q = queue.Queue()
        generator = iter(self.__shelf.keys())
        for key in generator:
            q.put(self.__shelf[key])
        return q

    def __len__(self):
        return self.size

    def __iter__(self):
        return self

    def __next__(self):
        if self.size:
            return self.get()
        raise StopIteration

    def __del__(self):
        if self.__shelf:
            self.__shelf.close()



import urllib
import logging
from typing import Text, Union
from ..exporters import Exporter, ExportResult
from ..utils import _validate_data_for_sql_table, \
                    _transform_data_for_sql_query, \
                    _describe_postgres_column
from ..exceptions import MissingConfigError, \
                                           InvalidConfigError, \
                                           MissingModuleError, \
                                           ExportError

logger = logging.getLogger(__name__)

try:
    import psycopg2
except ModuleNotFoundError:
    logger.error("""Module psycopg2 not available. Install using:
                    pip install resilient-exporters[postgres]""")
    raise

class PostgreSQLExporter(Exporter):
    """Exporter for PostgreSQL.

    Args:
        target_host (str):
        target_port (int):
        username (str):
        password (str):
        database (str):
        default_table (str):
        create_table_if_inexistent (bool): Default to False
        **kwargs : the keyword arguments to pass down to parent's class Exporter
    Raises:
        InvalidConfigError: if it cannot retrieve the server information, which
            is likely due an invalid configuration of the target.

    .. admonition:: Example

        .. code-block:: python

            from resilient_exporters.exporters import PostgreSQLExporter

            exporter = PostgreSQLExporter(target_host="myserver.domain.net",
                                          username="username",
                                          password="my-password",
                                          database="profiles",
                                          default_table="scientists")

            data = {"name": "Richard Feynman",
                    "age": 69}
            exporter.send(data)
    """
    def __init__(self,
                 conn_string: Text = None,
                 target_host: Text = None,
                 database: Text = None,
                 target_port: int = 5432,
                 username: Text = None,
                 password: Text = None,
                 default_table: Text = None,
                 **kwargs):
        assert conn_string or (target_host and target_port), "Either a connection string or target host is needed."
        super(PostgreSQLExporter, self).__init__(**kwargs)

        if conn_string:
            self.__conn = psycopg2.connect(conn_string)
        else:
            self.__conn = psycopg2.connect(host=target_host,
                                           port=port,
                                           user=username,
                                           password=password,
                                           dbname=database)
        self.__cur = self.__conn.cursor()
        self.__default_table = default_table

        # List tables in the database
        try:
            #self.__cur.execute("SELECT * FROM INFORMATION_SCHEMA.TABLES;")
            self.__cur.execute("SELECT * FROM pg_catalog.pg_tables WHERE schemaname != 'pg_catalog' \
                                AND schemaname != 'information_schema';")
            self.__all_tables = {}
            for row in self.__cur.fetchall():
                self.__all_tables[row[1]] = {}
        except Exception as e:
            logging.error(e)
            raise InvalidConfigError(self, "Cannot retrieve database info. Is \
                                            the configuration valid? Does the user have read rights?")
        
        # Get information on columns for each table
        for table_name in self.__all_tables.keys():
            self.__cur.execute(f"SELECT TABLE_NAME, COLUMN_NAME, DATA_TYPE, ORDINAL_POSITION, IS_NULLABLE, \
                                 CHARACTER_MAXIMUM_LENGTH, NUMERIC_PRECISION, DATETIME_PRECISION \
                                 FROM information_schema.columns WHERE table_name='{table_name}';")
            for col in self.__cur.fetchall():
                print(col)
                self.__all_tables[table_name][col[1]] = _describe_postgres_column(col)

    def send(self,
             data: Union[dict, tuple],
             table: Text = None) -> ExportResult:
        """Inserts data into a table. Reuses default database and
        tables names, if provided at initialisation.

        Args:
            data (Union[dict, tuple]): a dict or tuple representing the document to insert into the
                collection. If a dict, the keys must be the column names. If a tuple, there must be
                as many elements as there are columns in the table.
            table (str): name of the target table. If `None`, will use
                the default value set at initialisation. Default is `None`.

        Returns:
            ExportResult: the result in the form (ObjectId, True) if successful,
                (None, False) otherwise.

        Raises:
            MissingConfigError: if it cannot find a database and/or collection
                in the arguments and default values.
        """
        if table is None:
            if self.__default_table is None:
                    raise MissingConfigError(self, "No table given by argument nor default table configured.")
            table = self.__default_table

        if table in self.__all_tables.keys():
            # Validates the data, raises errors in case something is wrong
            _validate_data_for_sql_table(data, self.__all_tables[table])
            columns, values = _transform_data_for_sql_query(data)

            if isinstance(data, dict):
                print(f"INSERT INTO {table}({columns}) VALUES({values});")
                self.__cur.execute(f"INSERT INTO {table}({columns}) VALUES({values});", table)
            else:
                self.__cur.execute(f"INSERT INTO {table} VALUES({values});", table)
            success = bool(self.__cur.rowcount)
            self.__conn.commit()
            return ExportResult(None, success)
        else:
            raise InvalidConfigError(self, f"""Table {table} does not exist in
                                                database. Provide the name of 
                                                an existing table""")



from collections import namedtuple

ExportResult = namedtuple("ExportResult", ["returned", "successful"])

from .exporter import Exporter
from .exporter_pool import ExporterPool
from .exporter_file import FileExporter
from .exporter_mongodb import MongoDBExporter
from .exporter_elasticsearch import ElasticSearchExporter
from .exporter_postgresql import PostgreSQLExporter

__all__ = ["ExportResult", "Exporter", "ExporterPool", "FileExporter",
           "MongoDBExporter", "ElasticSearchExporter", "PostgreSQLExporter"]



import os
import orjson as json
import logging
import shutil
from typing import Optional, Text, Union, Iterable
from resilient_exporters.exporters import Exporter, ExportResult
from resilient_exporters.exceptions import ExportError

logger = logging.getLogger(__name__)

class FileExporter(Exporter):
    """Exporter for a text file.

    Args:
        target_file (str): the path of the file. The user must have write
            rights on the file.
        max_lines (int): the maximum number of lines in the file, incl. the
            lines already present. If `None`, there's no limit. Default to `None`.
        append (bool): if `True`, will append the data to the file.
            Otherwise, it will overwrite the file. Default to `True`.
        **kwargs : the keyword arguments to pass down to parent's class Exporter
    .. admonition:: Example

        .. code-block:: python

            import os
            from resilient_exporters.exporters import FileExporter

            exporter = FileExporter("local_file.txt",
                                    max_lines=1000,
                                    append=False,
                                    use_memory=False, # see exporters.Exporter
                                    save_unsent_data=True)

            data = {"name": "Richard Feynman",
                    "age": 69}
            exporter.send(data)
    """
    def __init__(self,
                 target_file: Text,
                 max_lines: Optional[int] = None,
                 append: bool = True,
                 **kwargs):
        super(FileExporter, self).__init__(**kwargs)

        #Create file if it doesn't exist
        if not os.path.isfile(target_file):
            open(target_file, "x").close()
        self.__filename = target_file
        self.__max_lines = max_lines #max lines in file
        self.__remaining_lines = -1

        self.start(append=append)

        # Some logic in case there's a max lines
        if self.__max_lines is not None:

            # Count number of lines in file
            tmp_count = len(self.target_file.readlines())

            # Alert if the maximum is already reached
            if tmp_count >= self.__max_lines:
                logger.warning(f"""there's already {tmp_count} lines in file,
                                you asked for a maximum of {self.__max_lines}.
                                Nothing will be added to the file.""")

            # Set remaining lines
            self.__remaining_lines = self.__max_lines - tmp_count

    @property
    def remaining_lines(self) -> int:
        """It is the amount of lines the file can still contain before reaching
        the limit passed at initialisation (cf `max_lines` keyword argument at
        init). Returns 0 if there's no limit or the limit has been reached."""
        return max(self.__remaining_lines, 0)

    def write_lines(self, data: Union[Text, dict]):
        if not isinstance(data, list):
            data = [data]
        for piece in data:
            if self.__remaining_lines != 0:
                line = piece if not isinstance(piece, dict) else json.dumps(piece)
                print(line, file=self.target_file)
                self.__remaining_lines -= 1
            else:
                logger.warning(f"Can't write more data in file {self.__filename}")

    """remove_lines:
       Removing by copying the file (except lines of given indices) into a
       new file, then replacing the old file with the new file. By default,
       removes the first line.
    """
    def remove_lines(self, indices: Iterable[int] = None) -> bool:
        if indices is None:
            indices = [0]
        new_filename = self.__filename + ".new"
        failed = False
        try:
            new_file = open(new_filename, "w")
            for count, line in enumerate(self.target_file):
                if count not in indices:
                    new_file.write(line)
        except Exception:
            failed = True
        finally:
            self.target_file.close()
            self.target_file = new_file

        if not failed:
            shutil.copyfile(new_filename, self.__filename)
            os.remove(new_filename)
            return not True
        return failed

    def send(self, data: Union[Text, dict]) -> ExportResult:
        """Writes the data into the file. Each call adds a new line in the
        file.

        Args:
            data (Union[Text, dict]): a string or `dict` with the data to
                write into the file. If a `dict`, it will be converted into a
                json document.

        Returns:
            ExportResult: if the operation was successful, returns (None, True),
                          otherwise (None, False)

        Raises:
            ExportError: if the IO stream is closed.
        """
        if self.target_file.closed:
            logger.error(f"IO stream closed for FileExporter {self.name}")
            raise ExportError(self, "Stream closed. \
                                     Has the FileExporter been stopped?")
        if self.__max_lines is None:
            self.write_lines(data)
        elif self.__remaining_lines > 0:
            self.write_lines(data)
        # @TODO a should new data replace old data if the file is full?
        #elif self.keep_new_data:
        #    self.remove_lines()
        #    self.write_lines(data)
        else:
            logger.warning(f"File {self.__filename} is full.")
            return ExportResult(None, False)
        return ExportResult(None, True)

    def start(self, append: bool = True):
        """Restarts the exporter by reopening an IO stream to the file. If there
        were no stream yet, it will create one.

        Args:
            append (bool): if True, opens the file in 'append' mode, else in
                "write" mode. Default is True.
        """
        try:
            if self.target_file.closed:
                logger.debug(f"Restarting the FileExporter {self.name}")
                self._create_stream_for_file(append)
        except AttributeError:
            # There's then no self.target_file and we should create it
            self._create_stream_for_file(append)

    def stop(self):
        """Stops the exporter, by closing the IO stream. The exporter must be
        stopped for another process to be able to read the file."""
        logger.debug(f"Stopping the FileExporter {self.name}")
        self.target_file.close()

    def _create_stream_for_file(self, append: bool = True):
        self.target_file = open(self.__filename, "a+" if append else "w+")

    def __del__(self):
        self.stop()


# -*- coding: utf-8 -*-

"""
resilient_exporters.exceptions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module contains the cutom exceptions used in resilient_exporters.
"""

__all__ = ["ResilientExporterException", "MissingModuleError", \
           "MissingConfigError", "InvalidConfigError", "ExportError", \
           "ConnectionError", "DataTypeError", "TimeoutError"]

class ResilientExporterException(Exception):
    """Something wrong occured when using exporters."""

    def __init__(self, exporter=None, *args, **kwargs):
        if exporter:
            self.type = type(exporter).__name__
            self.name = exporter.name
            self.content = f"Exception raised in {self.name} of type {self.type}"
            super(ResilientExporterException, self).__init__(self.content)
        else:
            super(ResilientExporterException, self).__init__(*args, **kwargs)

class MissingModuleError(ResilientExporterException, ModuleNotFoundError):
    """Module not found. Can be installed doing resilient-exporters[...]"""

class MissingConfigError(ResilientExporterException):
    """A piece of configuration is missing"""

class InvalidConfigError(ResilientExporterException):
    """A piece of configuration is wrong"""

class ExportError(ResilientExporterException):
    """An export job failed"""

class ConnectionError(ExportError):
    """Could not connect to target"""

class DataTypeError(ExportError, TypeError):
    """Provided is of the wrong type"""

class TimeoutError(ExportError):
    """An export attempt has timed out"""



import urllib
import logging
from typing import Text
from resilient_exporters.exporters import Exporter, ExportResult
from resilient_exporters.exceptions import MissingConfigError, \
                                           InvalidConfigError, \
                                           MissingModuleError

logger = logging.getLogger(__name__)

try:
    import pymongo
except ModuleNotFoundError:
    logger.error("""PyMongo not available. Install using:
                    pip install resilient-transmitter[mongo]""")
    raise MissingModuleError

class MongoDBExporter(Exporter):
    """Exporter for MongoDB.

    Args:
        target_ip (str):
        target_port (int):
        username (str):
        password (str):
        default_db (str):
        default_collection (str):
        **kwargs : the keyword arguments to pass down to parent's class Exporter
    Raises:
        InvalidConfigError: if it cannot retrieve the server information, which
            is likely due an invalid configuration of the target.

    .. admonition:: Example

        .. code-block:: python

            import os
            from resilient_exporters.exporters import MongoDBExporter

            exporter = MongoDBExporter(target_ip="127.0.0.1",
                                       username=os.environ["MONGO_USERNAME"],
                                       password=os.environ["MONGO_PASSWORD"],
                                       default_db="profiles",
                                       default_db="scientists")

            data = {"name": "Richard Feynman",
                    "age": 69}
            exporter.send(data)
    """

    def __init__(self,
                 target_ip: Text,
                 target_port: int = 27017,
                 username: Text = None,
                 password: Text = None,
                 default_db: Text = None,
                 default_collection: Text = None,
                 **kwargs):
        super(MongoDBExporter, self).__init__(**kwargs)

        self.target_ip = target_ip
        self.target_port = target_port
        self.username = username
        self.default_db = default_db
        self.default_collection = default_collection

        if self.target_ip is None:
            logger.error("No IP address provided.")
            raise ValueError("No IP address provided")

        # Parse username and password
        if username:
            username = urllib.parse.quote_plus(self.username)
        if password:
            password = urllib.parse.quote_plus(password)

        # Create kwargs for MongoClient
        if target_port:
            kwargs["port"] = self.target_port
        if username:
            kwargs["username"] = username
        if password:
            kwargs["password"] = password

        # Create MongoClient
        self.__client = pymongo.MongoClient(self.target_ip, **kwargs)

        # Test connection
        try:
            self.__client.server_info()
        except pymongo.errors.ServerSelectionTimeoutError as e:
            logging.debug(e)
            raise InvalidConfigError(self, "Cannot retrieve server info. Is \
                                            the target valid?")

        # Keeping default database/collection if provided
        self.db = None
        if self.default_db:
            self.db = self.__client[self.default_db]
        self.collection = None
        if self.default_collection:
            self.collection = self.db[self.default_collection]

    @property
    def client(self) -> pymongo.MongoClient:
        """The MongoDB client. It cannot be replaced."""
        return self.__client

    def send(self,
             data: dict,
             db: Text = None,
             collection: Text = None) -> ExportResult:
        """Inserts data into a collection. Reuses default database and
        collection names, if provided at initialisation.

        Args:
            data (dict): a dict representing the document to insert into the
                collection.
            db (str): name of the target database. If `None`, will use the
                default value. Default is `None`.
            collection (str): name of the target colleciton. If `None`, will use
                the default value. Default is `None`.

        Returns:
            ExportResult: the result in the form (ObjectId, True) if successful,
                (None, False) otherwise.

        Raises:
            MissingConfigError: if it cannot find a database and/or collection
                in the arguments and default values.
        """
        data = data.copy()
        if db is None and self.db is not None:
            if collection is None and self.collection is not None:
                try:
                    self.collection.insert_one(data)
                except Exception as e:
                    logger.error(e)
                    return ExportResult(False, None)
                return ExportResult(data["_id"], True)
            elif collection is not None:
                try:
                    self.db[collection].insert_one(data)
                except Exception as e:
                    logger.error(e)
                    return ExportResult(False, None)
                return ExportResult(data["_id"], True)
            else:
                raise MissingConfigError(self, "No reference to a database \
                                          and collection found.")
        elif db is not None and collection is not None:
            try:
                self.__client[db][collection].insert_one(data)
            except Exception as e:
                logger.error(e)
                return ExportResult(None, False)
            return ExportResult(data["_id"], True)
        return ExportResult(None, False)



import abc
import pathlib
import logging
import time
from collections.abc import MutableSequence, Set
from functools import wraps
from typing import Callable, Union, Optional, List, Text, Any
from resilient_exporters.utils import is_able_to_connect, _DataStore
from resilient_exporters.exporters import ExportResult

logger = logging.getLogger(__name__)

class Exporter(metaclass=abc.ABCMeta):
    """Base class of all exporters and ExporterPool. It includes a non-
    implemented `send` function that its children must implement. It has
    `abc.ABCMeta` as a metaclass to enforce the implementation of `send` by all
    the children classes.
    """

    TEST_URL: Optional[Text] = None
    """The URL to use to test a connection. If None, the URL will be
       "https://www.google.com".
    """

    __initialized: bool = False
    __instantiated: int = 0

    def __init__(self,
                 transform: Optional[Callable] = None,
                 timeout: int = 30, # in seconds
                 use_memory: bool = True,
                 manual_reexport: bool = True,
                 *,
                 tmp_file: Union[Text, pathlib.Path, None] = None,
                 #@TODO: reinitialize_tmp_file: bool = True,
                 save_unsent_data: bool = True,
                 #@TODO: name: Optional[Text] = None,
                 test_url: Optional[Text] = None):
        if timeout < 0:
            raise ValueError("timeout must be non-negative and non null")

        self.__transform = transform
        self.timeout = timeout
        self.manual_reexport = manual_reexport
        self.TEST_URL = test_url

        self.tmp_filename = tmp_file if tmp_file is None else tmp_file
        self.name = f"exporter_{Exporter.__instantiated}"
        self._run_transform = transform is not None
        self._save_unsent_data = save_unsent_data

        self._only_stop_if_all_data_sent = False
        self._datastore = _DataStore(use_memory=use_memory,
                                     shelf_filename=self.tmp_filename)
        self.__approx_qsize = 0
        self.__initialized = True
        Exporter.__instantiated += 1
        self.__is_sending_unsent_data = False

        # If __init__ is called by an exporter that implements the `send` method
        # which means the type of self must be != "Exporter", e.g."FileExporter"
        # we wrap the `send` method to include pre- and post- processes.
        if type(self).__name__ != "Exporter":
            self.send = self._exporter_wrapper(self.send)
        logger.debug(f"Exporter {self.name} instantiated.")

    @property
    def use_memory(self) -> bool:
        """The value `use_memory` of the instance.

        When set to ``True``, it loads the data into memory if it was previously
        using a file (previous value was ``False``), or vice versa.

        Args:
            new_val (bool): new value for `use_memory`.

        Returns:
            bool: its value
        """
        return self._datastore.use_memory

    @use_memory.setter
    def use_memory(self, new_val: bool) -> bool:
        self._datastore.use_memory = new_val
        return new_val

    #@TODO: getter for transform
    #@property
    #def transform(self) -> Optional[Callable]:
    #    """Getter for `transform`. It has no corresponding setter. `transform`
    #    cannot be changed after instanciation to avoid corrupting previously
    #    transformed data.
    #
    #    Returns:
    #        Optional[Callable]: the `transform` function or `None` if none has
    #        been provided at initialisation.
    #    """
    #    return self.__transform

    def transform(self, data: Any) -> Any:
        """Applies `transform` to the given data.

        Args:
            data (Any): the data to pass to `transform`.

        Returns:
            Any: the output of `transform`

        Raises:
            Exception: if the output of `transform(data)` is None.
        """
        if self.__transform is None:
            return data
        logger.debug(f"transform input data of type {type(data)}")
        data = self.__transform(data.copy())
        if data is None:
            raise Exception("Empty data returned by transform.")
        logger.debug(f"transform output data of type {type(data)}")
        return data

    def save_unsent_data(self,
                         data: Any,
                         kwargs: dict,
                         exporter_name: str) -> None:
        """Saves the data in memory or disk, depending on the value of
        `use_memory` of the instance, so it can be sent later.

        Args:
            data (Any): the core data.
            kwargs (dict): the keyword arguments to pass to `send` for the given
                           data.
            exporter_name (str): the name of the exporter.

        Returns:
            None
        """
        data_to_keep = {"data": data,
                        "kwargs": kwargs,
                        "exporter": exporter_name}
        self._datastore.put(data_to_keep)

    def has_unsent_data(self) -> bool:
        """Assesses if there's saved, unsent data.

        Returns:
            bool: `True` if there's saved, unsent data. `False` otherwise.
        """
        return self._datastore.size > 0

    def _exporter_wrapper(self, send_func: Callable) -> Callable:
        """Wrapper for the implemented ``send`` method. It checks data type,
        runs the call ``transform`` on the data, and processes the result.

        Args:
            send_func (Callable): an implementation of ``Exporter.send``

        Returns:
            Callable: the wrapped ``send`` method.
        """

        @wraps(send_func)
        def wrapper(data: Any, *args, **kwargs) -> Union[ExportResult, \
                                                         List[ExportResult]]:
            logger.debug(f"Sending input data of type {type(data)}")
            if data is None:
                raise ValueError("No data.")
            if isinstance(data, MutableSequence):
                raise ValueError("Data cannot be a list or a MutableSequence")
            if isinstance(data, Set):
                raise ValueError("Data cannot be a tuple or a Set")
            if self._run_transform:
                data = self.transform(data)
            logger.debug(f"Calling `send` of exporter {self.name}")
            result = send_func(data, **kwargs)
            logger.debug(f"Processing result {result}")
            return self._process_result(result, data, kwargs)
        return wrapper

    def send_unsent_data(self) -> List[ExportResult]:
        """Tries to send the previously saved, unsent data.

        Returns:
            List[ExportResult]: list of the results of the export jobs.
        """
        self.__is_sending_unsent_data = True
        results = [self.send(d["data"], **d["kwargs"]) \
                   for d in self._datastore]
        self.__is_sending_unsent_data = False
        return results

    @abc.abstractmethod
    def send(self, data, **kwargs) -> Union[ExportResult, List[ExportResult]]:
        """Abstract method to be implemented by children.

        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def _process_result(self,
                        result: ExportResult,
                        data: Any,
                        kwargs: dict) -> Union[ExportResult, \
                                               List[ExportResult]]:
        if result.successful:
            if not self.manual_reexport \
                and is_able_to_connect(self.TEST_URL)\
                and self.has_unsent_data():
                logger.info("Attempt to send previously unsent data.")
                return [result] + self.send_unsent_data()
        elif not result.successful and self._save_unsent_data:
            self.save_unsent_data(data, kwargs, self.name)
        return result

    def _replace_datastore(self, new_datastore: _DataStore):
        self._datastore = new_datastore

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'send') and
                callable(subclass.send) and
                hasattr(subclass, 'has_unsent_data') and
                callable(subclass.has_unsent_data))

    def __setattr__(self, attr, val):
        if self.__initialized and attr in ('transform', 'tmp_file'):
            raise ValueError(f'{attr} attribute should not be set after '
                             f'{self.__class__.__name__} is initialized.')
        super(Exporter, self).__setattr__(attr, val)

    def __del__(self):
        if self._only_stop_if_all_data_sent and self.has_unsent_data():
            while self.has_unsent_data():
                self.send_unsent_data()
                time.sleep(0.2)
        del self



import logging
from typing import Optional, Text, Iterable
from resilient_exporters.exporters import Exporter, ExportResult
from resilient_exporters.exceptions import MissingModuleError, \
                                           MissingConfigError, \
                                           InvalidConfigError

logger = logging.getLogger(__name__)

try:
    import elasticsearch
except ModuleNotFoundError:
    logger.error("""Elasticsearch not available. Install using:
                    pip install resilient-transmitter[elastic]""")
    raise MissingModuleError

class ElasticSearchExporter(Exporter):
    """Exporter for ElasticSearch.

    Args:
        target_ip (str): an IP address of a ElasticSearch server.
        target_port (int): the port to connect to. Default to 9300.
        username (str): the username for authentication.
        password (str): the password as plain text for authentication.
                        Use an environement variable for security.
        cluster_hosts (Iterable[Text]): cluster of hosts, passed to ES's client
                        application.
        cloud_id (str): cloud id used to connect to a Elastic Cloud server.
                        A username and password is most likely required to be
                        able to connect.
        api_key (str): a base64 encoded token to authenticate to a ElasticSearch
                       server.
        sniff_on_start (bool): see Elasticsearch documentation.
        default_index (str): a default index to use when ``send`` is called. If
                        None, an index will have to be provided as an argument
                        when calling ``send``.
        **kwargs : the keyword arguments to pass down to parent's class Exporter

    .. admonition:: Warning

        If ``target_ip`` is provided, it will supercede ``cluster_hosts``.
    """

    def __init__(self,
                 target_ip: Text = None,
                 target_port: int = 9300,
                 username: Text = None,
                 password: Text = None,
                 cluster_hosts: Iterable[Text] = None,
                 cloud_id: Text = None,
                 api_key: Text = None, #base 64 encoded token
                 sniff_on_start: bool = True,
                 default_index: Optional[Text] = None,
                 use_ssl: bool = False,
                 ssl_certfile: Text = None,
                 ssl_ca_certs: Text = None,
                 **kwargs):
        super(ElasticSearchExporter, self).__init__(**kwargs)

        self.target_ip = target_ip
        self.target_port = target_port
        self.cluster_hosts = cluster_hosts
        self.cloud_id = cloud_id
        self.api_key = api_key
        self.sniff_on_start = sniff_on_start
        self.default_index = default_index
        self.use_ssl = use_ssl
        self.ssl_certfile = ssl_certfile
        self.ssl_ca_certs = ssl_ca_certs

        # Need to provide an address
        if self.target_ip is None \
           and self.cluster_hosts is None \
           and self.cloud_id is None:
            logger.error("No target address provided.")
            raise ValueError

        kwargs = {}
        if username and password:
            kwargs["http_auth"] = (username, password)
            #kwargs["scheme"] = "https"
        if self.cloud_id:
            kwargs["cloud_id"] = self.cloud_id
        if self.api_key:
            kwargs["api_key"] = self.api_key
        if self.sniff_on_start:
            kwargs["sniff_on_start"] = self.sniff_on_start
        if self.use_ssl:
            kwargs["use_ssl"] = self.use_ssl
        if self.ssl_certfile:
            kwargs["client_cert"] = self.ssl_certfile
        if self.ssl_ca_certs:
            kwargs["ca_certs"] = self.ssl_ca_certs

        hosts = None
        if self.target_ip:
            hosts = {"host": self.target_ip}
            if self.target_port:
                hosts["port"] = self.target_port
        elif self.cluster_hosts:
            hosts = self.cluster_hosts

        if hosts:
            self.__client = elasticsearch.Elasticsearch(hosts, **kwargs)
        else:
            self.__client = elasticsearch.Elasticsearch(**kwargs)

    @property
    def client(self) -> elasticsearch.Elasticsearch:
        """The Elasticsearch client. It cannot be replaced."""
        return self.__client

    def send(self, data: dict, index: Optional[Text] = None) -> ExportResult:
        """Indexes the data into an ElasicSearch index.

        Args:
            data (dict): the data, as a dict, to index.
            index (str): the index name. If `None`, it uses the default value
                provided at initialisation.

        Returns:
            ExportResult: (Object, True) if successful, (None, False) otherwise.

        Raises:
            MissingConfigError: if no index is found.
            InvalidConfigError: if cannot send data because of a configuration
                issue (authentication or other type of issues).
        """
        if index is None:
            index = self.default_index
        if index is None:
            raise MissingConfigError(self, "No index found.")
        try:
            res = self.__client.index(index=index, body=data)
            return ExportResult(res, True)
        except elasticsearch.exceptions.ConnectionError:
            logger.warning("elasticsearch.exceptions.ConnectionError")
        except elasticsearch.exceptions.RequestError:
            logger.error("elasticsearch.exceptions.RequestError")
            raise InvalidConfigError(self, "elasticsearch.exceptions.RequestError")
        return ExportResult(None, False)



import urllib
import logging
from typing import Text
from resilient_exporters import utils
from resilient_exporters.exporters import Exporter, ExportResult
from resilient_exporters.exceptions import MissingConfigError, \
                                           InvalidConfigError, \
                                           MissingModuleError, \
                                           ExportError

logger = logging.getLogger(__name__)

try:
    import _mssql
except ModuleNotFoundError:
    logger.error("""Module pymssql not available. Install using:
                    pip install resilient-exporters[sqlserver]""")
    raise
except ImportError as e:
    logger.error(e)
    logger.error("""Is FreeTDS installed in the system?""")
    raise

class SQLServerExporter(Exporter):
    """Exporter for Microsoft SQL Server, Azure SQL DB and MySQL.

    Args:
        target_host (str):
        target_port (int):
        username (str):
        password (str):
        database (str):
        default_table (str):
        create_table_if_inexistent (bool):
        **kwargs : the keyword arguments to pass down to parent's class Exporter
    Raises:
        InvalidConfigError: if it cannot retrieve the server information, which
            is likely due an invalid configuration of the target.

    .. admonition:: Example

        .. code-block:: python

            import os
            from resilient_exporters.exporters import SQLServerExporter

            exporter = SQLServerExporter(target_host="myserver.domain.net",
                                         username="username",
                                         password="my-password",
                                         database="profiles",
                                         default_table="scientists")

            data = {"name": "Richard Feynman",
                    "age": 69}
            exporter.send(data)
    """
    def __init__(self,
                 target_host: Text,
                 database: Text,
                 target_port: int = 1433,
                 username: Text = None,
                 password: Text = None,
                 default_table: Text = None,
                 **kwargs):
        super(SQLServerExporter, self).__init__(**kwargs)

        self.__conn = _mssql.connect(server=target_host, 
                                     user=username, 
                                     password=password, 
                                     database=database)

        self.__default_table = default_table

        # List tables in the database
        try:
            self.__conn.execute_query("SELECT * FROM INFORMATION_SCHEMA.TABLES")
            self.__all_tables = {}
            for row in self.__conn:
                self.__all_tables[row["TABLE_NAME"]] = {}
        except _mssql.MSSQLDataBaseException as e:
            logging.error(e)
            raise InvalidConfigError(self, "Cannot retrieve database info. Is \
                                            the configuration valid? Does the user have read rights?")
        
        # Get information on columns for each table
        for table_name in self.__all_tables.keys():
            self.__conn.execute_query(f"""SELECT COLUMN_NAME, DATA_TYPE, ORDINAL_POSITION, IS_NULLABLE, CHARACTER_MAXIMUM_LENGTH, NUMERIC_PRECISION, DATETIME_PRECISION FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME='{table_name}'""")
            for col in self.__conn:
                if col["CHARACTER_MAXIMUM_LENGTH"]:
                    precision = col["CHARACTER_MAXIMUM_LENGTH"]
                elif col["NUMERIC_PRECISION"]:
                    precision = col["NUMERIC_PRECISION"]
                else:
                    precison = col["DATETIME_PRECISION"]

                col_info = (col["DATA_TYPE"], col["ORDINAL_POSITION"], col["IS_NULLABLE"], precision)

                self.__all_tables[table_name][col["COLUMN_NAME"]] = col_info

    def send(self,
             data: dict,
             table: Text = None) -> ExportResult:
        """Inserts data into a table. Reuses default database and
        tables names, if provided at initialisation.

        Args:
            data (dict): a dict representing the document to insert into the
                collection.
            table (str): name of the target table. If `None`, will use
                the default value. Default is `None`.

        Returns:
            ExportResult: the result in the form (ObjectId, True) if successful,
                (None, False) otherwise.

        Raises:
            MissingConfigError: if it cannot find a database and/or collection
                in the arguments and default values.
        """
        if table is None:
            if self.__default_table is None:
                    logger.error("""No table name given and no default table 
                                    configured.""")
                    raise MissingConfigError(self, """No table given by 
                            argument nor default table configured. Please 
                            provide a table name.""")
            table = self.__default_table

        if table in self.__all_tables.keys():
            # Validates the data, raises errors in case something is wrong
            utils.validate_data_for_sql_table(data, self.__all_tables[table])
            
            columns = ",".join(data.keys())
            values = ",".join([ f"'{val}'" if isinstance(val, str) else str \
                                                (val) for val in data.values()])

            self.__conn.execute_non_query(f"""INSERT INTO {table}({columns})
                                              VALUES({values})""", table)
            return ExportResult(None, True)
        else:
            logger.error(f"Invalid table name provided: {table}")
            raise InvalidConfigError(self, f"""Table {table} does not exist in
                                                database. Provide the name of 
                                                an existing table""")
                                            


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['resilient_exporters',
 'resilient_exporters.exceptions',
 'resilient_exporters.exporters']

package_data = \
{'': ['*']}

install_requires = \
['orjson', 'requests']

extras_require = \
{'all': ['pymongo>=3,<4',
         'elasticsearch[async]>6.0',
         'psycopg2-binary>=2.7,<3.0'],
 'elastic': ['elasticsearch[async]>6.0'],
 'mongo': ['pymongo>=3,<4'],
 'postgres': ['psycopg2-binary>=2.7,<3.0']}

setup_kwargs = {
    'name': 'resilient-exporters',
    'version': '0.1.3',
    'description': 'A package to export data to databases resiliently.',
    'long_description': '# Resilient Exporters\n![PyPI](https://img.shields.io/pypi/v/resilient-exporters?logo=pypi&logoColor=white&style=for-the-badge)\n![GitHub Build Status](https://img.shields.io/github/workflow/status/arbfay/resilient-exporters/Python%20package?logo=github&style=for-the-badge)\n![License](https://img.shields.io/github/license/arbfay/resilient-exporters?style=for-the-badge)\n![Python Version](https://img.shields.io/badge/3.6+%20-%2314354C.svg?label=PYTHON&style=for-the-badge&logo=python&logoColor=white)\n\nResilient-exporters abstracts away common tasks when sending or saving data from an application. It has been designed to send data to different targets and manage common issues for applications running on edge devices such as a Raspberry Pi or Nvidia Jetson Nano:\n- Internet connection interruptions;\n- Highly variable frequency of data transfers;\n\nIf a connection is lost, it automatically saves the data and retries later when the connection is recovered and when a new request to send data is made. To avoid consuming too much memory or disk space, it has a specific configurable flush.\n\nIf an application wants to send more data than is momentally manageable, it multiplies the transmission jobs (using multithreading, if available) and saves the data (queuing), to avoid back-pressure and reducing the application\'s speed.\n\nOf course, it can break if:\n- the data to transmit is _almost always_ more important than the available bandwidth;\n- the interruptions are too long compared to the available memory or disk space;\n\nWe have designed it particularly for a Raspberry Pi 3B+ device running a Linux distribution.\n\n## Installation\nTo use the package:\n```\npip install resilient-exporters\n```\n### Dev installation\nIf you\'d like to have a editable, up-to-date version of the files, do:\n```\ngit clone https://github.com/arbfay/resilient-exporters.git && \\\npip install -e resilient-exporters/ && \\\npip install -r resilient-exporters/dev_requirements.txt\n```\n\n## Usage\nCurrently supported:\n- Text file\n- MongoDB\n- ElasticSearch\n- PostgreSQL\n\nSome features for some exporters might be missing. Raise an issue on Github to ask for an implementation and help improve the package.\n\n### Store in a file\n```python\nfrom resilient_exporters import FileExporter\n\nexporter = FileExporter(target_file="mydata.txt",\n                        max_lines=100)\n\nmydata = ["value1", "value2"]\nexporter.send(mydata)\n```\n\n### To MongoDB\n```python\nfrom resilient_exporters import MongoDBExporter\n\nexporter = MongoDBExporter(target_ip="127.0.0.1",\n                           target_port=27017,\n                           username="username",\n                           password="password",\n                           default_db="my_db",\n                           default_collection="my_collection")\n\nmydata = {"field1": "value1"}\nexporter.send(mydata)\n```\n\n### To ElasticSearch\n```python\nfrom resilient_exporters import ElasticSearchExporter\n\nexporter = ElasticSearchExporter(target_ip="127.0.0.1",\n                                 default_index="my_index",\n                                 use_ssl=True,\n                                 ssl_certfile="/path/to/file",\n                                 sniff_on_start=True)\n\nmydata = {"field1": "value1"}\nexporter.send(mydata)\n```\n\n# To PostgreSQL\n```python\nfrom resilient_exporters.exporters import PostgreSQLExporter\n\nexporter = PostgreSQLExporter(target_host="myserver.domain.net",\n                              username="username",\n                              password="my-password",\n                              database="profiles",\n                              default_table="scientists")\n\ndata = {"name": "Richard Feynman",\n        "age": 69}\nexporter.send(data)\n```\n\n### Multiple distant targets - Pools\nEdge devices are more and more powerful, and are capable of managing multiple distant targets without much overhead thanks to `resilient-exporters`. If you\'re taking advantage of this, you might need sometimes to replicate data across different databases of the same type (e.g. NoSQL, document-based databases). However, if you use multiple exporters, all the features will be duplicated and can generate inefficiencies (multiple temporary files, multiple queues, etc.).\n\nInstead, use `resilient_exporters.ExporterPool` which pools exporters _and_ other pools to expose only one `send` method for all the exporters and to ensure a more efficient management of the resources. To use it:\n```python\nfrom resilient_exporters import ExporterPool\nfrom resilient_exporters import MongoDBExporter, ElasticSearchExporter\n\nexporters = [\n  MongoDBExporter(target_ip="127.0.1.10",\n                  target_port=1234,\n                  default_db="my_db",\n                  default_collection="my_collection"),\n  ElasticSearchExporter(cloud_id="cloud id",\n                        api_key="api key",\n                        default_index="my_index")]\n\npool = ExporterPool(exporters, use_memory=False)\n\nmydata = {"metric": 2}\npool.send(mydata)\n```\n\n## Transform data before sending\nTo transform data before it gets sent by an exporter or a pool, one can add a function that takes the input data and returns the transformed data:\n```python\nfrom resilient_exporters import MongoDBExporter\n\ndef transform(data):\n  data["metric"] = (data["metric"] / 2) * .5\n  return data\n\nexporter = MongoDBExporter(target_ip="127.0.1.10",\n                           target_port=1234,\n                           default_db="my_db",\n                           default_collection="my_collection",\n                           transform=transform)\n\nmydata = {"metric": 2}\nexporter.send(mydata)\n```\n>**NOTE**: it can also be passed to a pool with the same key argument `tranform` at initialisation. When doing so, transform functions of individual exporters will be superseded by the pool\'s transform function.\n\n## Additional information\nThe `resilient_exporters.Exporter` is at the core of the package. All the other exporters inherits from it.\n\n`Exporter` manages the export of data to a target, however each target need specific logic to send data. All these subclasses, such as `FileExporter` or `MongoDBExporter`, implements the `Exporter.send` method and manages the needed options. Some exporters might need additional packages to be usable:\n- `pymongo` for `MongoDBExporter`\n- `elasticsearch` for `ElasticSearchExporter`\n- `psycopg2` for `PostgreSQLExporter`\n\n## Documentation\nMore documentation available [here.](https://resilient-exporters.readthedocs.io)\n\n## Suggestions and contribution\nPlease open a GitHub issue for bugs or feature requests.\nContact the contributors for participating.\n',
    'author': 'Fayçal Arbai',
    'author_email': 'arbai.faycal@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/arbfay/resilient-exporters.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)



import pathlib
import logging
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, Union, Optional, List, Iterable, Text, Any
from resilient_exporters.exporters import Exporter, ExportResult
from resilient_exporters.exceptions import InvalidConfigError
from resilient_exporters.utils import is_able_to_connect

logger = logging.getLogger(__name__)

class ExporterPool(Exporter):
    """Enables pooling of exporters for improved efficiency and performance.
    All the exporters will be managed by the pool, including the saving of
    unsent data, with only one ``send`` call to be used. It also offers a
    multithreading option to run the exporters' ``send`` functions in parallel,
    to speed up the execution of ``ExporterPool.send``.

    Args:
        exporters (Iterable[Exporter]): a list of
            ``resilient_exporters.exporters.Exporter``.
        transform (Callable): a function to be invoked at each ``send`` call on
            the passed data. It must return data or exceptions will be raised.
        num_threads (int): the number of threads to use for ``send`` calls. Must
            be greater than 1. If 1, multithreading is disabled. Default to 1.
        wait_for_result (bool): value to decide if the instance has to wait for
            the results of ``send`` calls or not when multithreading is enabled.
        manual_reexport (bool): if True, the user is responsible to call the
            function ``send_unsent_data`` when appropriate. If False, the
            instance will manage that automatically by assessing the necessity
            to call the function at each ``send`` call; the criterias are 1)
            there's unsent data, 2) there's an Internet connection. Default to
            False.

    .. admonition:: Note

        If using parallelism, the default behaviour is to wait for the results
        of all calls. One can disable this behaviour with `wait_for_result` set
        to ``False``, and ``ExporterPool.send`` will be then non-blocking, but
        will return ``None``.

    Raises:
        InvalidConfigError: if ``num_threads`` is < 1.

    .. admonition:: Example

        .. code-block:: python

            import resilient_exporters as rex

            exporter1 = rex.exporters.FileExporter("local_file.txt")
            exporter2 = rex.exporters.FileExporter("/path/to/network/file.txt")
            pool = ExporterPool([exporter1, exporter2], num_threads=2)

            line = "A string to be written in a file"
            pool.send(line)

    Attributes:
        num_threads (int): number of threads used by the instance.
        wait_for_result (bool): value to decide if the instance has to wait for
            the results of ``send`` calls or not when multithreading is enabled.
    """
    __futures = []
    __instantiated = 0

    def __init__(self,
                 exporters: Optional[Iterable[Exporter]],
                 transform: Optional[Callable] = None,
                 num_threads: int = 1,
                 wait_for_result: bool = True,
                 use_memory: bool = True,
                 manual_reexport: bool = False,
                 *,
                 tmp_file: Union[Text, pathlib.Path, None] = None,
                 save_unsent_data: bool = True):
                 #@TODO: name: Optional[Text] = None):
        super(ExporterPool, self).__init__(transform=transform,
                                           use_memory=use_memory,
                                           tmp_file=tmp_file,
                                           manual_reexport=manual_reexport,
                                           save_unsent_data=save_unsent_data)
        if num_threads < 1:
            raise InvalidConfigError(self, "num_threads must be >= 1; use \
                                    num_threads=1 to disable multithreading.")

        self.num_threads = num_threads
        self.wait_for_result = wait_for_result

        self.__exporters = {}
        if exporters is not None:
            for exporter in exporters:
                exporter._replace_datastore(self._datastore)
                if self._run_transform:
                    exporter._run_transform = False
                exporter._save_unsent_data = not self._save_unsent_data
                self.__exporters[exporter.name] = exporter

        self.__instantiated += 1
        self.name = f"exporterpool_{ExporterPool.__instantiated}"

    @property
    def exporters(self) -> dict:
        """A dictionary of the contained exporters."""
        return self.__exporters

    def add_exporter(self, exporter: Exporter) -> None:
        """Adds an exporter to the pool. Use this function to add an exporter
        after the pool has been initialised. It removes the responsability to
        run the `transform` method from the exporter, and to save unsent data.

        Args:
            exporter (Exporter): an exporter.
        """
        exporter.use_memory = self.use_memory
        if self._run_transform:
            exporter._run_transform = False
        exporter._save_unsent_data = not self.save_unsent_data
        self.__exporters[exporter.name] = exporter

    def send(self, data: Any, **kwargs) -> List[ExportResult]:
        """Runs the `send` method of all its exporters. If the pool's
        `num_threads` attribute is > 1, it will execute all the calls in
        separate threads.

        .. admonition:: Note
            The key arguments passed at the call of the method will be passed
            down to all the exporters. Make sure they all have different
            keywords.

        Args:
            data (Any): the data to export.
            **kwargs (Any): the keyword arguments to pass down to the exporters'
                `send` methods.

        Returns:
            List[ExportResult]: a list of the exporters' results.
        """
        results = []
        if self.num_threads <= 1:
            for exporter in self:
                results.append(exporter.send(data, **kwargs))
        else:
            # use multithreading
            futures = []
            with ThreadPoolExecutor(max_workers=self.num_threads) as executor:
                for exporter in self:
                    future = executor.submit(exporter.send, data, **kwargs)
                    futures.append(future)

                    def save_if_failed(future):
                        if future.result() == False:
                            self.save_unsent_data(data, kwargs, exporter.name)

                    future.add_done_callback(save_if_failed)
            if self.wait_for_result:
                results += [f.result() for f in futures]
            else:
                self.__futures += futures
        return results

    def _process_result(self,
                        results: Iterable[ExportResult],
                        data: Any,
                        kwargs: dict) -> Union[bool, List[bool]]:
        summed_res = sum([r.successful for r in results])
        if summed_res == len(results):
            # all expeditions have been successful
            if not self.manual_reexport \
                and is_able_to_connect(self.TEST_URL) \
                and self.has_unsent_data():
                logger.info("Attempt to send previously unsent data.")
                return results + self.send_unsent_data()
        elif summed_res == 0 and self._save_unsent_data:
            # all have failed
            self.save_unsent_data(data, kwargs, self.name)
        else:
            #mixed results
            for exporter, res in zip(self, results):
                if not res.successful and self._save_unsent_data:
                    self.save_unsent_data(data, kwargs, exporter.name)
        return results

    def send_unsent_data(self) -> List[ExportResult]:
        """Tries to send the previously saved, unsent data.

        Returns:
            List[ExportResult]: list of the results of the export jobs.
        """
        self.__is_sending_unsent_data = True
        results = [self.send(d["data"], d["exporter"], **d["kwargs"]) \
                   for d in self._datastore]
        self.__is_sending_unsent_data = False
        return results

    def __len__(self) -> int:
        return len(self.__exporters)

    def __iter__(self):
        self.__iterator_count = 0
        self.__exporters_as_list = list(self.exporters.values())
        return self

    def __next__(self) -> Optional[Exporter]:
        if self.__iterator_count < len(self):
            res = self.__exporters_as_list[self.__iterator_count]
            self.__iterator_count += 1
            return res
        del self.__exporters_as_list
        raise StopIteration

    def __del__(self):
        if not self.wait_for_result:
            res = [f.result() for f in self.__futures]
            del res
            del self.__futures
